// Copyright 2014 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package runtime

import "unsafe"

// This file contains the implementation of Go channels
// and select statements.

// Outline of the synchronization protocols.
//
// Single channel sync send/recv first check the opposite wait queue (WaitQ).
// If the queue is not empty, it dequeues a waiter (SudoG), copies the data
// directly to/from the waiter buffer and wakes up the waiter.
// If the queue is empty, it queues itself into wait queue and parks. When it is
// woken up, the operation is already executed, so it just returns.
// SudoG.signal is set to nil before parking and is left intact during wake up.
// Close sets SudoG.signal to CLOSED, in such case woken up goroutine acts accordingly.
//
// Single channel async send/recv first check whether the operation can be executed
// (buffer is not full/empty, respectively). If the operation can be executed,
// it copies data to/from the buffer, adjusts indices and wakes up an opposite
// waiter if there is any. If the operation cannot be executed at the moment,
// it queues itself into wait queue and parks. When it is woken up, it retries
// from the beginning.
// SudoG.signal is set to nil before parking and is left intact during wake up.
// Close sets SudoG.signal to CLOSED, in such case woken up goroutine
// accomplishes the operation w/o retry.
//
// Select operations proceed in 3 passes. In the pass 1 channels are checked one-by-one;
// if some channel is ready, the operation is executed (as outlined above for
// single channel operations) and select returns. In the pass 2 rediness of the
// channels is re-checked and select queues itself as waiter on all channels.
// If no channel turned out to be ready in pass 2, select blocks. In the pass 3
// (after wake up) select dequeues itself from unsuccessful channels. If it is
// woken up by an operation on a sync channel or close, select returns. Otherwise
// it retries from the beginning.
// SudoG.signal for all channels is setup to point to a single signal variable,
// signal variable is set to PREWAIT before enqueue. The signal variable is operated
// with CAS and allows to resolve competition between concurrent wakers and the
// select goroutine. When SudoG is removed from a wait queue, SudoG.signal is set to nil;
// this allows select to skip dequeueing it again. Below is the signal state machine:
//
// PREWAIT -> WAIT: select commits to park
//
// PREWAIT -> sg|CLOSED: sg chan is closed, select is not woken up
// PREWAIT -> READY: an async chan is ready or select discovers a ready chan during pass 2, select is not woken up
// PREWAIT -> SYNC: a sync chan is ready but the operation is not yet completed, select is not woken up
//
// WAIT -> sg|CLOSED: the same as above, but select is woken up
// WAIT -> READY:
// WAIT -> SYNC:
//
// SYNC -> sg: operation on the sync chan is completed now
const (
	sgPrewait = iota
	sgClosed
	sgWait
	sgReady
	sgSync
)

const (
	maxAlign  = 8
	hchanSize = unsafe.Sizeof(hchan{})
	debugChan = false
)

// TODO: make hchan.buf an unsafe.Pointer, not a *uint8

func makechan(t *chantype, size int64) *hchan {
	elem := t.elem

	// compiler checks this but be safe.
	if elem.size >= 1<<16 {
		gothrow("makechan: invalid channel element type")
	}
	if hchanSize%maxAlign != 0 || elem.align > maxAlign {
		gothrow("makechan: bad alignment")
	}
	if size < 0 || int64(uintptr(size)) != size || (elem.size > 0 && uintptr(size) > (maxMem-hchanSize)/uintptr(elem.size)) {
		panic("makechan: size out of range")
	}

	var c *hchan
	if elem.kind&kindNoPointers != 0 || size == 0 {
		// allocate memory in one call
		c = (*hchan)(gomallocgc(hchanSize+uintptr(size)*uintptr(elem.size), nil, flagNoScan))
		if size > 0 && elem.size != 0 {
			c.buf = (*uint8)(add(unsafe.Pointer(c), hchanSize))
		} else {
			c.buf = (*uint8)(unsafe.Pointer(c)) // race detector uses this location for synchronization
		}
	} else {
		c = new(hchan)
		c.buf = (*uint8)(newarray(elem, uintptr(size)))
	}
	c.elemsize = uint16(elem.size)
	c.elemtype = elem
	c.dataqsiz = uint(size)

	if debugChan {
		println("makechan: chan=", c, "; elemsize=", elem.size, "; elemalg=", elem.alg, "; dataqsiz=", size)
	}
	return c
}

// chanbuf(c, i) is pointer to the i'th slot in the buffer.
func chanbuf(c *hchan, i uint) unsafe.Pointer {
	return add(unsafe.Pointer(c.buf), uintptr(i)*uintptr(c.elemsize))
}

// entry point for c <- x from compiled code
//go:nosplit
func chansend1(t *chantype, c *hchan, elem unsafe.Pointer) {
	chansend(t, c, elem, true, gogetcallerpc(unsafe.Pointer(&t)))
}

/*
 * generic single channel send/recv
 * If block is not nil,
 * then the protocol will not
 * sleep but return if it could
 * not complete.
 *
 * sleep can wake up with g.param == nil
 * when a channel involved in the sleep has
 * been closed.  it is easiest to loop and re-run
 * the operation; we'll see that it's now closed.
 */
func chansend(t *chantype, c *hchan, ep unsafe.Pointer, block bool, callerpc uintptr) bool {
	if raceenabled {
		fn := chansend
		pc := **(**uintptr)(unsafe.Pointer(&fn))
		raceReadObjectPC(t.elem, ep, callerpc, pc)
	}

	if c == nil {
		if !block {
			return false
		}
		gopark(nil, nil, "chan send (nil chan)")
		return false // not reached
	}

	if debugChan {
		println("chansend: chan=", c)
	}

	if raceenabled {
		fn := chansend
		pc := **(**uintptr)(unsafe.Pointer(&fn))
		racereadpc(unsafe.Pointer(c), pc, callerpc)
	}

	// Fast path: check for failed non-blocking operation without acquiring the lock.
	//
	// After observing that the channel is not closed, we observe that the channel is
	// not ready for sending. Each of these observations is a single word-sized read
	// (first c.closed and second c.recvq.first or c.qcount depending on kind of channel).
	// Because a closed channel cannot transition from 'ready for sending' to
	// 'not ready for sending', even if the channel is closed between the two observations,
	// they imply a moment between the two when the channel was both not yet closed
	// and not ready for sending. We behave as if we observed the channel at that moment,
	// and report that the send cannot proceed.
	//
	// It is okay if the reads are reordered here: if we observe that the channel is not
	// ready for sending and then observe that it is not closed, that implies that the
	// channel wasn't closed during the first observation.
	if !block && c.closed == 0 && ((c.dataqsiz == 0 && c.recvq.first == nil) ||
		(c.dataqsiz > 0 && c.qcount == c.dataqsiz)) {
		return false
	}

	var t0 int64
	if blockprofilerate > 0 {
		t0 = gocputicks()
	}

	golock(&c.lock)
	if c.closed != 0 {
		gounlock(&c.lock)
		panic("send on closed channel")
	}

	if c.dataqsiz == 0 { // synchronous channel
		sg, wake := c.recvq.dequeue(sgSync)
		if sg != nil { // found a waiting receiver
			if raceenabled {
				racesync(c, sg)
			}
			// Disable preemption. If we communicate with a select,
			// it can be still running. It such case completion of the operation
			// is the store to sg->signal below. We don't want this goroutine
			// to be preempted until that store. Otherwise the select would
			// need to spin in while(signal == SYNC) using a mix of osyield and gosched.
			mp := acquirem()
			gounlock(&c.lock)

			if sg.elem != nil {
				memmove(unsafe.Pointer(sg.elem), ep, uintptr(c.elemsize))
			}
			if signalp := sg.signal; signalp != nil {
				sg.signal = nil
				goatomicstorep(unsafe.Pointer(signalp), unsafe.Pointer(sg))
			}
			if wake {
				if sg.releasetime != 0 {
					// Yes, this is ugly.  On 64-bit sg.releasetime has type
					// int.  On 32-bit it has type int64.  There's no easy way
					// to assign to both types in Go.  At some point we'll
					// write the Go types directly instead of generating them
					// via the C types.  At that point, this nastiness goes away.
					*(*int64)(unsafe.Pointer(&sg.releasetime)) = gocputicks()
				}
				goready(sg.g)
			}
			releasem(mp)
			return true
		}

		if !block {
			gounlock(&c.lock)
			return false
		}

		// no receiver available: block on this channel.
		gp := getg()
		mysg := acquireSudog()
		if t0 != 0 {
			mysg.releasetime = -1
		}
		mysg.elem = (*uint8)(ep)
		mysg.waitlink = nil
		gp.waiting = mysg
		mysg.g = gp
		mysg.signal = nil
		gp.param = nil
		c.sendq.enqueue(mysg)
		goparkunlock(&c.lock, "chan send")

		// someone woke us up.
		if mysg.signal != nil {
			if c.closed == 0 {
				gothrow("chansend: spurious wakeup")
			}
			panic("send on closed channel")
		}
		if mysg.releasetime > 0 {
			goblockevent(int64(mysg.releasetime)-t0, 3)
		}
		if mysg != gp.waiting {
			gothrow("G waiting list is corrupted!")
		}
		gp.waiting = nil
		releaseSudog(mysg)
		return true
	}

	// asynchronous channel
	// wait for some space to write our data
	var t1 int64
	for c.qcount >= c.dataqsiz {
		if !block {
			gounlock(&c.lock)
			return false
		}
		gp := getg()
		mysg := acquireSudog()
		if t0 != 0 {
			mysg.releasetime = -1
		}
		mysg.g = gp
		mysg.elem = nil
		mysg.signal = nil
		c.sendq.enqueue(mysg)
		goparkunlock(&c.lock, "chan send")

		// someone woke us up - try again
		if mysg.releasetime != 0 {
			t1 = int64(mysg.releasetime)
		}
		releaseSudog(mysg)
		golock(&c.lock)
		if c.closed != 0 {
			gounlock(&c.lock)
			panic("send on closed channel")
		}
	}

	// write our data into the channel buffer
	if raceenabled {
		raceacquire(chanbuf(c, c.sendx))
		racerelease(chanbuf(c, c.sendx))
	}
	memmove(chanbuf(c, c.sendx), ep, uintptr(c.elemsize))
	c.sendx++
	if c.sendx == c.dataqsiz {
		c.sendx = 0
	}
	c.qcount++

	// wake up a waiting receiver
	sg, wake := c.recvq.dequeue(sgReady)
	gounlock(&c.lock)
	if wake {
		if sg.releasetime != 0 {
			*(*int64)(unsafe.Pointer(&sg.releasetime)) = gocputicks()
		}
		goready(sg.g)
	}
	if t1 > 0 {
		goblockevent(t1-t0, 3)
	}
	return true
}

func (q *waitq) enqueue(sgp *sudog) {
	sgp.link = nil
	if q.first == nil {
		q.first = sgp
		q.last = sgp
		return
	}
	q.last.link = sgp
	q.last = sgp
}

func (q *waitq) dequeue(what int) (*sudog, bool) {
loop:
	for {
		sgp := q.first
		if sgp == nil {
			return nil, false
		}
		q.first = sgp.link
		if q.last == sgp {
			q.last = nil
		}

		if sgp.signal == nil {
			// single chan op
			if what == sgClosed {
				sgp.signal = (**sudog)(unsafe.Pointer(uintptr(sgClosed)))
			}
			return sgp, true
		}

		// select
		if sgp.g == getg() {
			// self select, not interesting
			sgp.signal = nil
			continue
		}
		for {
			old := *sgp.signal
			if old != sgState(sgPrewait) && old != sgState(sgWait) {
				// already signaled on a different chan
				sgp.signal = nil
				continue loop
			}
			new := sgState(what)
			if what == sgClosed {
				new = (*sudog)(unsafe.Pointer(uintptr(unsafe.Pointer(sgp)) | uintptr(sgClosed)))
			}
			// claim the right to signal this select
			if gocasp(unsafe.Pointer(sgp.signal), unsafe.Pointer(old), unsafe.Pointer(new)) {
				if what != sgSync {
					sgp.signal = nil
				}
				return sgp, old == sgState(sgWait) // do not wake if it was not waiting
			}
		}
	}
}

func sgState(x int) *sudog {
	return (*sudog)(unsafe.Pointer(uintptr(x)))
}

func racesync(c *hchan, sg *sudog) {
	racerelease(chanbuf(c, 0))
	raceacquireg(sg.g, chanbuf(c, 0))
	racereleaseg(sg.g, chanbuf(c, 0))
	raceacquire(chanbuf(c, 0))
}
